# app.py
# Author: Labeller <https://github.com/mdbloice/Labeller>
from flask_bootstrap import Bootstrap
from flask import Flask, render_template, g, request, abort, jsonify
import os
import glob
import random
import sqlite3
import pickle

app = Flask(__name__)
Bootstrap(app)

# Read the config file which contains the class names.
with open('labeller.pkl', 'rb') as to_read:
    class_names = pickle.load(to_read)

# Global
randomised = False

# SQLite Database
DATABASE = os.path.join('.', 'db', 'labels.db')
ROOT_IMAGE_PATH = os.path.join('.', 'static', 'images')

def get_db():
    db = getattr(g, '_database', None)

    if db is None:
        db = g._database = sqlite3.connect(DATABASE)
    return db

@app.teardown_appcontext
def teardown_db(exception):
    db = getattr(g, '_database', None)

    if db is not None:
        db.close()

def query_db(query, args=(), one=False):
    cur = get_db().execute(query, args)
    rv = cur.fetchall()
    cur.close()
    return (rv[0] if rv else None) if one else rv

@app.route('/api/image', methods=['POST'])
def set_image():

    if not request.json or not 'image' in request.json or not 'label' in request.json or not 'label_string' in request.json:
        abort(400)

    q = "INSERT INTO labels (image, label, label_string) VALUES ('%s', %s, '%s')"  \
        % (request.json['image'].split('images/')[-1], request.json['label'], request.json['label_string'])

    con =  get_db()
    cursor = con.cursor()
    cursor.execute(q)
    con.commit()
    cursor.close()

    return jsonify({"Status": "OK"})

@app.route('/')
def index():
        # Perform all image searching here in the index rather than
        # outside of any of the routes
        extensions = ["*.png", "*.PNG", "*.jpg", "*.JPG", "*.jpeg", "*.JPEG"]
        images = []

        for ext in extensions:
            images.extend(glob.glob(os.path.join(ROOT_IMAGE_PATH, ext)))

        total_n_images = len(images)  # Total number of images labelled or unlabelled

        images_to_remove = []

        for image in images:

            image = str.split(image, os.path.join(ROOT_IMAGE_PATH, os.sep))[-1]
            con = get_db()
            cursor = con.cursor()
            cursor.execute('SELECT EXISTS(SELECT 1 FROM labels WHERE image="%s" LIMIT 1);' % image)
            data=cursor.fetchall()

            if data[0][0] == 1:
                images_to_remove.append(os.path.join(ROOT_IMAGE_PATH, image))

            cursor.close()

        print("Length of images to remove: %s" % len(images_to_remove))

        # This is happening for every refresh and page visit.
        images = list(set(images) - set(images_to_remove))

        # Randomise the order of the images, only once if not done before.
        global randomised
        if not randomised:
            print("Randomising image list...")
            random.shuffle(images)
            randomised = True

        remaining_image_count = len(images)

        return render_template(
        'index.html', images=images, remaining_image_count=remaining_image_count, total_n_images=total_n_images)

@app.route('/about.html')
def about():
    return render_template('about.html')

@app.route('/labels.html')
def labels():

    # Get all the labels in the database
    labels = query_db("SELECT * FROM labels")

    class_distribution = {}
    global class_names

    # Get the frequency of each class name
    for c in class_names:
        r = query_db('SELECT count(*) FROM labels WHERE label_string="%s"' % c)
        class_distribution[c] = r[0][0]

    return render_template('labels.html', labels=labels, class_distribution=class_distribution)



# htmlElements.py
# Author: Marcus D. Bloice <https://github.com/mdbloice>
# Contains dynamic HTML elements that cannot be copied from the resources
# directory, where static HTML and CSS is stored.
import pkg_resources
import os


class Button():
    def __init__(self, btn_id, btn_label) -> None:
        self.btn_label = btn_label
    def get_html(self):
        return '<a id="%s" class="btn btn-lg btn-default" role="button" style="width: 100px">%s</a>' % (self.btn_id, self.btn_label)


class KeyPressJS():
    def __init__(self, classes) -> None:
        self.classes = classes
    def get_html(self):

        js_elements = []

        l = len(self.classes) if len(self.classes) <= 9 else 10

        for i in range(l):

            js = """
            if (e.keyCode == %s) {
                console.log(
                    "Handler for keypress fired with keyCode: " + e.keyCode + " (%s)"
                );
            postToDB(%s);
            getNewImage();

            """ % ((i+1)+48, self.classes[i], i+1)

            js_elements.append(js)

        return """
        $(document).keypress(function (e) {

            %s

            }
        });
        """ % ' '.join(js_elements)


# Navbar is not currently used as we use the static navbar.html file
# from the package's resources directory.
class Navbar():
    def __init__(self) -> None:
        pass

    def get_html(self) -> str:

        navbar_contents = """
        {% block navbar %}
        <nav class="navbar navbar-default">
            <div class="container-fluid">
                <div class="navbar-header">
                <a class="navbar-brand" href="/"><span class="glyphicon glyphicon-tag"></span> Labeller</a>
                </div>
                <ul class="nav navbar-nav">
                <li class="{{ 'active' if active_page == 'index' else '' }}"><a href="/"><span class="glyphicon glyphicon-home"></span> Label Images</a></li>
                <li class="{{ 'active' if active_page == 'labels' else '' }}"><a href="labels.html"><span class="glyphicon glyphicon-oil"></span> Label Database</a></li>
                <li class="{{ 'active' if active_page == 'about' else '' }}"><a href="about.html"><span class="glyphicon glyphicon-question-sign"></span> About</a></li>
                </ul>
            </div>
        </nav>
        {% endblock %}
        """

        return navbar_contents


class Footer():
    def __init__(self, footerText='Created with <a href="https://github.com/mdbloice/Labeller" target="_blank"><span class="glyphicon glyphicon-tag"></span> Labeller</a>') -> None:
        self.footerText = footerText

    def get_html(self) -> str:
        return """
        <div class="container">
        <hr>
        <footer class="bg-light text-center text-lg-start" style="padding: 10px;">
            <div class="text-center p-3">
            %s
            </div>
        </footer>
        </div>
        """ % self.footerText


class Index():
    def __init__(self, class_names) -> None:
        self.class_names = class_names
        self.n_classes = len(class_names)

    def get_html(self):

        keyboard_shortcuts = []
        keyboard_shortcut_js = []
        buttons = []
        button_js = []

        for i in range(self.n_classes):
            keyboard_shortcuts.append("<li><kbd>%s</kbd> for <b>%s</b></li>" % (i+1, self.class_names[i]))
            keyboard_shortcut_js.append('case "%s": postToDB(%s, "%s"); getNewImage(); break;' % (i+1, i, self.class_names[i]))
            buttons.append('<a id="%s" class="btn btn-lg btn-primary" role="button">%s</a>' % (self.class_names[i], self.class_names[i]))
            button_js.append('$("#%s").click(function () { postToDB(%s, "%s"); getNewImage(); });' % (self.class_names[i], i, self.class_names[i]))

        # Use only the first 10 classes and bind to the keys 1, 2, 3, 4, 5, 6, 7, 8, 9, and 0
        if self.n_classes >= 10:
            keyboard_shortcuts = keyboard_shortcuts[:9]
            keyboard_shortcuts.append("<li><kbd>%s</kbd> for <b>%s</b></li>" % (0, self.class_names[9]))
            keyboard_shortcut_js = keyboard_shortcut_js[:9]
            keyboard_shortcut_js.append('case "%s": postToDB(%s, "%s"); getNewImage(); break;' % (0, 9, self.class_names[9]))

        index_html = pkg_resources.resource_string(__name__, os.path.join('resources', 'index.html')).decode("utf-8")

        return index_html % (' '.join(keyboard_shortcut_js), ' '.join(button_js), ' '.join(keyboard_shortcuts), ' '.join(buttons))



# createLabeller.py
# Author: Marcus D. Bloice <https://github.com/mdbloice>
# Licensed under the terms of the MIT Licence.
import argparse
import sys
import sqlite3
import os
from collections import OrderedDict
import glob
import pkg_resources
import pickle

# Internal imports
from labeller.htmlElements import Footer, Index

parser = argparse.ArgumentParser(description='Generate an image labelling \
    application.', prog='python -m labeller')
parser.add_argument('class_names', metavar='class_names', nargs='+',
    help='a list of class names seperated by spaces')

args = parser.parse_args()

n_classes = len(args.class_names)

if n_classes <= 1:
    print('Minimum number of classes is 2, you provided %s: %s' % (n_classes,
        args.class_names))
    sys.exit(1)

# If we get here the arguments look good, so we can generated the app.
print("Attempting to generate application with %s classes: %s." % (n_classes,
    ', '.join(args.class_names)))

extensions = ["*.png", "*.PNG", "*.jpg", "*.JPG", "*.jpeg", "*.JPEG"]

im_dir = os.path.join('static', 'images')

if not os.path.exists(os.path.join('.', im_dir)):
    print("No directory named '%s' found in the current directory: %s). Aborting." % (im_dir, os.getcwd()))
    sys.exit(1)
else:
    print("Found a directory named %s: %s\nSearching for images with the following extensions: %s" % (im_dir, os.path.join(os.getcwd(), im_dir), ' '.join(extensions)))

image_paths = []
for ext in extensions:
    image_paths.extend(glob.glob(os.path.join('.', im_dir, ext)))

if len(image_paths) <= 0:
    print("No images found in %s. Aborting." % os.path.join(os.getcwd(), im_dir))
    sys.exit(1)
else:
    print("Found %s images in %s" %(len(image_paths), os.path.join(os.getcwd(), im_dir)))

# Create a config file as a pickle containing the class names.
with open(os.path.join('.', 'labeller.pkl'), 'wb') as to_write:
    pickle.dump(args.class_names, to_write)

# Create a database using the class labels provided
# We create the database in createLabeller.py and not in app.py as we
# wish to create the database only on generation of the app,
# so that if we wish to use the labeller -m command to create an
# app it can be used to also reset the database.
os.makedirs(os.path.join('.', 'db'), exist_ok=True)
conn = sqlite3.connect(os.path.join('.', 'db', 'labels.db'))
conn.execute('CREATE TABLE IF NOT EXISTS labels (id INTEGER PRIMARY KEY, image STRING, label INTEGER, label_string STRING)')
conn.execute('CREATE INDEX IF NOT EXISTS im_name_index ON labels(image)')

already_labelled = 0

# This needs to be checked carefully for compatibility with Windows OSes
# due to paths and path seperators.
for image_path in image_paths:
    im = str.split(image_path, os.path.join("static", "images", os.sep))[-1]
    cursor = conn.cursor()
    res = cursor.execute('SELECT EXISTS(SELECT 1 FROM labels WHERE image="%s" LIMIT 1);' % im)
    data=cursor.fetchall()
    if data[0][0]:
        already_labelled += 1
    cursor.close()

print("%s images have already been labelled." % already_labelled)

conn.close()

print("Generating application with %s classes with the labels %s for %s images." % (n_classes, ', '.join(args.class_names), len(image_paths)))

# Create directory structure.
# We serve images from static/images, but could be changed with 'send_from_directory'
os.makedirs(os.path.join('.', 'templates'), exist_ok=True)
os.makedirs(os.path.join('.', 'static', 'styles'), exist_ok=True)
os.makedirs(os.path.join('.', 'static', 'images'), exist_ok=True)

# Use pkg_resources to open files or data installed with the package
app_py = pkg_resources.resource_string(__name__, os.path.join('resources', 'app.py'))
dashboard_css = pkg_resources.resource_string(__name__, os.path.join('resources', 'dashboard.css'))
about_html = pkg_resources.resource_string(__name__, os.path.join('resources', 'about.html'))
labels_html = pkg_resources.resource_string(__name__, os.path.join('resources', 'labels.html'))
navbar_html = pkg_resources.resource_string(__name__, os.path.join('resources', 'navbar.html'))
favicon_ico = pkg_resources.resource_string(__name__, os.path.join('resources', 'favicon.ico'))

# Copy any STATIC files to their appropriate directories here
with open(os.path.join('.', 'app.py'), 'wb') as to_write:
    to_write.write(app_py)

with open(os.path.join('.', 'static', 'styles', 'dashboard.css'), 'wb') as to_write:
    to_write.write(dashboard_css)

with open(os.path.join('.', 'static', 'favicon.ico'), 'wb') as to_write:
    to_write.write(favicon_ico)

with open(os.path.join('.', 'templates', 'about.html'), 'wb') as to_write:
    to_write.write(about_html)

with open(os.path.join('.', 'templates', 'labels.html'), 'wb') as to_write:
    to_write.write(labels_html)

with open(os.path.join('.', 'templates', 'navbar.html'), 'wb') as to_write:
    to_write.write(navbar_html)

# Create any DYNAMIC content here.
footer = Footer()
with open(os.path.join('.', 'templates', 'footer.html'), 'w') as to_write:
    to_write.write(footer.get_html())

index = Index(class_names=args.class_names)
with open(os.path.join('.', 'templates', 'index.html'), 'w') as to_write:
    to_write.write(index.get_html())
